@@ -1,5 +1,5 @@
 /*
-  Copyright(C) 2014-2016 Brazil
+  Copyright(C) 2014-2018 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -23,6 +23,8 @@
 
 #include "groonga/accessor.h"
 #include "groonga/array.h"
+#include "groonga/arrow.h"
+#include "groonga/cache.h"
 #include "groonga/column.h"
 #include "groonga/config.h"
 #include "groonga/dat.h"
@@ -37,10 +39,12 @@
 #include "groonga/ii.h"
 #include "groonga/obj.h"
 #include "groonga/operator.h"
+#include "groonga/option.h"
 #include "groonga/output.h"
 #include "groonga/pat.h"
 #include "groonga/request_canceler.h"
 #include "groonga/request_timer.h"
+#include "groonga/string.h"
 #include "groonga/table.h"
 #include "groonga/thread.h"
 #include "groonga/time.h"
@@ -49,3 +53,4 @@
 #include "groonga/window_function.h"
 #include "groonga/windows.h"
 #include "groonga/windows_event_logger.h"
+#include "groonga/vector.h"
