@@ -1,6 +1,6 @@
 /* -*- c-basic-offset: 2 -*- */
 /*
-  Copyright(C) 2009-2014 Brazil
+  Copyright(C) 2009-2017 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -39,7 +39,7 @@ grn_token_cursor_open_initialize_token_filters(grn_ctx *ctx,
     return;
   }
 
-  token_cursor->token_filter.data = GRN_MALLOCN(void *, n_token_filters);
+  token_cursor->token_filter.data = GRN_CALLOC(sizeof(void *) * n_token_filters);
   if (!token_cursor->token_filter.data) {
     return;
   }
@@ -63,11 +63,10 @@ grn_token_cursor_open(grn_ctx *ctx, grn_obj *table,
   grn_token_cursor *token_cursor;
   grn_encoding encoding;
   grn_obj *tokenizer;
-  grn_obj *normalizer;
   grn_obj *token_filters;
   grn_table_flags table_flags;
   if (grn_table_get_info(ctx, table, &table_flags, &encoding, &tokenizer,
-                         &normalizer, &token_filters)) {
+                         NULL, &token_filters)) {
     return NULL;
   }
   if (!(token_cursor = GRN_MALLOC(sizeof(grn_token_cursor)))) { return NULL; }
@@ -76,6 +75,7 @@ grn_token_cursor_open(grn_ctx *ctx, grn_obj *table,
   token_cursor->encoding = encoding;
   token_cursor->tokenizer = tokenizer;
   token_cursor->token_filter.objects = token_filters;
+  token_cursor->token_filter.data = NULL;
   token_cursor->orig = (const unsigned char *)str;
   token_cursor->orig_blen = str_len;
   token_cursor->curr = NULL;
@@ -108,7 +108,7 @@ grn_token_cursor_open(grn_ctx *ctx, grn_obj *table,
   } else {
     int nflags = 0;
     token_cursor->nstr = grn_string_open_(ctx, str, str_len,
-                                          normalizer,
+                                          table,
                                           nflags,
                                           token_cursor->encoding);
     if (token_cursor->nstr) {
@@ -341,6 +341,10 @@ grn_token_cursor_close_token_filters(grn_ctx *ctx,
   grn_obj *token_filters = token_cursor->token_filter.objects;
   unsigned int i, n_token_filters;
 
+  if (!token_cursor->token_filter.data) {
+    return;
+  }
+
   if (token_filters) {
     n_token_filters = GRN_BULK_VSIZE(token_filters) / sizeof(grn_obj *);
   } else {
