@@ -16,6 +16,7 @@ module Groonga
                  "output_columns",
                  "use_range_index",
                  "post_filter",
+                 "sort_keys",
                ])
 
       def run_body(input)
@@ -73,18 +74,23 @@ module Groonga
         key << "#{input[:output_columns]}\0"
         key << "#{input[:use_range_index]}\0"
         key << "#{input[:post_filter]}\0"
+        key << "#{input[:sort_keys]}\0"
         dynamic_columns = DynamicColumns.parse(input)
         key << dynamic_columns.cache_key
         key
       end
 
       class ExecuteContext
+        include KeysParsable
+
         attr_reader :use_range_index
         attr_reader :enumerator
         attr_reader :order
         attr_reader :filter
         attr_reader :offset
         attr_reader :limit
+        attr_reader :post_filter
+        attr_reader :sort_keys
         attr_reader :dynamic_columns
         attr_accessor :current_offset
         attr_accessor :current_limit
@@ -92,7 +98,6 @@ module Groonga
         attr_reader :unsorted_result_sets
         attr_reader :temporary_tables
         attr_reader :threshold
-        attr_reader :post_filter
         def initialize(input)
           @input = input
           @use_range_index = parse_use_range_index(@input[:use_range_index])
@@ -101,6 +106,9 @@ module Groonga
           @filter = @input[:filter]
           @offset = (@input[:offset] || 0).to_i
           @limit = (@input[:limit] || 10).to_i
+          @post_filter = @input[:post_filter]
+          @sort_keys = parse_keys(@input[:sort_keys])
+
           @dynamic_columns = DynamicColumns.parse(@input)
 
           @current_offset = @offset
@@ -112,8 +120,6 @@ module Groonga
           @temporary_tables = []
 
           @threshold = compute_threshold
-
-          @post_filter = @input[:post_filter]
         end
 
         def close
@@ -183,7 +189,7 @@ module Groonga
             first_shard ||= shard
             shard_executor = ShardExecutor.new(@context, shard, shard_range)
             shard_executor.execute
-            break if @context.current_limit == 0
+            break if @context.current_limit.zero?
           end
           if first_shard.nil?
             message =
@@ -206,6 +212,240 @@ module Groonga
         end
       end
 
+      class Window
+        include Comparable
+
+        attr_reader :unit
+        attr_reader :step
+        def initialize(context, shard, shard_range, unit, step)
+          @context = context
+          @shard = shard
+          @shard_range = shard_range
+          @unit = unit
+          @step = step
+
+          @between = Groonga::Context.instance["between"]
+
+          @target_range = @context.enumerator.target_range
+
+          case @unit
+          when :day
+            @step_second = (60 * 60 * 24) * @step
+          when :hour
+            @step_second = (60 * 60) * @step
+          else
+            raise InvalidArgument "Unexpected unit: #{@unit.inspect}"
+          end
+        end
+
+        def <=>(other)
+          case @unit
+          when other.unit
+            @step <=> other.unit
+          when :day # @unit == :day && other.unit == :hour
+            1
+          else # @unit == :hour && other.unit == :day
+            -1
+          end
+        end
+
+        def each(table)
+          min = create_min_edge
+          max = create_max_edge(min)
+          shard_key = table.find_column(@context.enumerator.shard_key_name)
+          begin
+            current_min = min
+            while current_min < max do
+              next_min = compute_next_min_edge(current_min)
+              if next_min > max
+                current_max = max
+              else
+                current_max = WindowEdge.new(next_min.year,
+                                             next_min.month,
+                                             next_min.day,
+                                             next_min.hour,
+                                             next_min.minute,
+                                             next_min.second,
+                                             next_min.microsecond,
+                                             :exclude)
+              end
+              windowed_table = select_by_range(table,
+                                               shard_key,
+                                               current_min,
+                                               current_max)
+              if windowed_table.empty?
+                windowed_table.close
+              else
+                message = "[logical_range_filter][window] "
+                message << "<#{@shard.table_name}>: "
+                message << inspect_range(current_min, current_max)
+                Context.instance.logger.log(Logger::Level::DEBUG,
+                                            __FILE__,
+                                            __LINE__,
+                                            __method__.to_s,
+                                            message)
+                yield(windowed_table)
+              end
+              current_min = next_min
+            end
+          ensure
+            shard_key.close
+          end
+        end
+
+        private
+        def create_min_edge
+          min = @target_range.min
+          if min
+            year = min.year
+            month = min.month
+            day = min.day
+            hour = min.hour
+            minute = min.min
+            second = min.sec
+            microsecond = min.usec
+            border = @target_range.min_border
+          else
+            year = @shard_range.year
+            month = @shard_range.month
+            day = 1
+            hour = 0
+            minute = 0
+            second = 0
+            microsecond = 0
+            border = :include
+          end
+          WindowEdge.new(year, month, day, hour, minute, second, microsecond,
+                         border)
+        end
+
+        def create_max_edge(min)
+          max = @target_range.max
+          if max
+            year = max.year
+            month = max.month
+            day = max.day
+            hour = max.hour
+            minute = max.min
+            second = max.sec
+            microsecond = max.usec
+            border = @target_range.max_border
+          else
+            next_shard_edge = @shard_range.least_over_time
+            year = next_shard_edge.year
+            month = next_shard_edge.month
+            day = next_shard_edge.day
+            hour = 0
+            minute = 0
+            second = 0
+            microsecond = 0
+            border = :exclude
+          end
+          WindowEdge.new(year, month, day, hour, minute, second, microsecond,
+                         border)
+        end
+
+        def compute_next_min_edge(current_min)
+          next_edge = Time.at(current_min.to_time.to_i + @step_second)
+          year = next_edge.year
+          month = next_edge.month
+          day = next_edge.day
+          if @unit == :day
+            hour = 0
+          else
+            hour = next_edge.hour
+          end
+          minute = 0
+          second = 0
+          microsecond = 0
+          border = :include
+          WindowEdge.new(year, month, day, hour, minute, second, microsecond,
+                         border)
+        end
+
+        def select_by_range(table, shard_key, min, max)
+          expression = Expression.create(table)
+          begin
+            expression.append_object(@between, Operator::PUSH, 1)
+            expression.append_object(shard_key, Operator::PUSH, 1)
+            expression.append_operator(Operator::GET_VALUE, 1)
+            expression.append_constant(min.to_s, Operator::PUSH, 1)
+            expression.append_constant(min.border, Operator::PUSH, 1)
+            expression.append_constant(max.to_s, Operator::PUSH, 1)
+            expression.append_constant(max.border, Operator::PUSH, 1)
+            expression.append_operator(Operator::CALL, 5)
+            table.select(expression)
+          ensure
+            expression.close
+          end
+        end
+
+        def inspect_range(min, max)
+          range = ""
+          if min.border == :include
+            range << "["
+          else
+            range << "("
+          end
+          range << min.to_s
+          range << ","
+          range << max.to_s
+          if max.border == :include
+            range << "]"
+          else
+            range << ")"
+          end
+          range
+        end
+      end
+
+      class WindowEdge
+        include Comparable
+
+        attr_reader :year
+        attr_reader :month
+        attr_reader :day
+        attr_reader :hour
+        attr_reader :minute
+        attr_reader :second
+        attr_reader :microsecond
+        attr_reader :border
+        def initialize(year, month, day, hour, minute, second, microsecond,
+                       border)
+          @year = year
+          @month = month
+          @day = day
+          @hour = hour
+          @minute = minute
+          @second = second
+          @microsecond = microsecond
+          @border = border
+        end
+
+        def to_s
+          format = "%04d/%02d/%02d %02d:%02d:%02d"
+          format_values = [@year, @month, @day, @hour, @minute, @second]
+          unless @microsecond.zero?
+            format << ".%06d"
+            format_values << @microsecond
+          end
+          format % format_values
+        end
+
+        def to_time
+          Time.local(@year, @month, @day, @hour, @minute, @second, @microsecond)
+        end
+
+        def values
+          [@year, @month, @day, @hour, @minute, @second, @microsecond]
+        end
+
+        def <=>(other)
+          (values + [@border == :include ? 1 : 0]) <=>
+            (other.values + [other.border == :include ? 1 : 0])
+        end
+      end
+
       class ShardExecutor
         def initialize(context, shard, shard_range)
           @context = context
@@ -326,6 +566,11 @@ module Groonga
             return decide_use_range_index(false, reason, __LINE__, __method__)
           end
 
+          unless @context.sort_keys.empty?
+            reason = "sort_keys is used"
+            return decide_use_range_index(false, reason, __LINE__, __method__)
+          end
+
           current_limit = @context.current_limit
           if current_limit < 0
             reason = "limit is negative: <#{current_limit}>"
@@ -512,7 +757,7 @@ module Groonga
                               nil, nil,
                               nil, nil)
             else
-              sort_result_set(table)
+              add_result_set(table)
             end
           else
             if range_index
@@ -649,7 +894,7 @@ module Groonga
           create_expression(table) do |expression|
             yield(expression)
             result_set = table.select(expression)
-            sort_result_set(result_set)
+            add_result_set(result_set)
           end
         end
 
@@ -660,7 +905,7 @@ module Groonga
           end
         end
 
-        def sort_result_set(result_set)
+        def add_result_set(result_set)
           @context.temporary_tables.delete(result_set)
 
           if result_set.empty?
@@ -668,6 +913,18 @@ module Groonga
             return
           end
 
+          window = detect_window
+          if window
+            window.each(result_set) do |windowed_result_set|
+              sort_result_set(windowed_result_set)
+              break if @context.current_limit.zero?
+            end
+          else
+            sort_result_set(result_set)
+          end
+        end
+
+        def sort_result_set(result_set)
           @context.dynamic_columns.each_filtered do |dynamic_column|
             if result_set == @shard.table
               result_set = result_set.select_all
@@ -688,12 +945,25 @@ module Groonga
           end
 
           @unsorted_result_sets << result_set if result_set.temporary?
-          sort_keys = [
-            {
-              :key => @context.enumerator.shard_key_name,
-              :order => @context.order,
-            },
-          ]
+          if @context.sort_keys.empty?
+            sort_keys = [
+              {
+                :key => @context.enumerator.shard_key_name,
+                :order => @context.order,
+              },
+            ]
+          else
+            sort_keys = @context.sort_keys.collect do |sort_key|
+              if sort_key.start_with?("-")
+                key = sort_key[1..-1]
+                order = :descending
+              else
+                key = sort_key
+                order = :ascending
+              end
+              {:key => key, :order => order}
+            end
+          end
           if @context.current_limit > 0
             limit = @context.current_limit
           else
@@ -710,6 +980,37 @@ module Groonga
             @context.current_limit -= sorted_result_set.size
           end
         end
+
+        def detect_window
+          window_group_keys = []
+          @context.dynamic_columns.each_filtered do |dynamic_column|
+            window_group_keys.concat(dynamic_column.window_group_keys)
+          end
+          return nil if window_group_keys.empty?
+
+          # TODO: return nil if result_set is small enough
+
+          windows = []
+          @context.dynamic_columns.each do |dynamic_column|
+            next unless window_group_keys.include?(dynamic_column.label)
+            case dynamic_column.value.strip
+            when /\Atime_classify_(.+?)\s*\(\s*([a-zA-Z\d_]+)/
+              type = $1
+              column = $2
+              next if column != @context.enumerator.shard_key_name
+
+              case type
+              when "minute", "second"
+                windows << Window.new(@context, @shard, @shard_range, :hour, 1)
+              when "day", "hour"
+                unless @shard_range.is_a?(LogicalEnumerator::DayShardRange)
+                  windows << Window.new(@context, @shard, @shard_range, :day, 1)
+                end
+              end
+            end
+          end
+          windows.max
+        end
       end
     end
   end
